# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import gevent.monkey
import multiprocessing
import itertools,time,pdb
from . accumulators import accumulators_pool
from ... utilities.mlog import MLog
GLOBAL_ORDER_ID = 0

class GeneticMutationFactors(object):
    
    def __init__(self, del_prob, add_prob, change_prob, conver_prob,
                 cover_prob=0.05, generation=500, group_num=10, 
                 parallel = None,objective=None):
        self._del_prob = del_prob # 删除原始特征概率
        self._add_prob = add_prob # 添加新特征概率
        self._change_prob =  change_prob #变异概率
        self._conver_prob = conver_prob # 收敛差值
        self._cover_prob = cover_prob # 变异种群覆盖率
        self._generations = generation
        self._group_num = group_num
        self._objective = objective
        self._parallel = parallel if parallel is not None else multiprocessing.cpu_count()
        
    # 初代种群生成
    def ga_generate_ori(self, ori_field, add_field):
        ## 选取原始特征过程
        ori_list = (np.random.uniform(0, 1, (1, len(ori_field))) > self._del_prob).tolist()[0]
        add_list = (np.random.uniform(0, 1, (1, len(add_field))) < self._add_prob).tolist()[0]
        new_list = ori_list + add_list
        new_list = [1 if i> 0 else 0 for i in new_list]
        return np.array(new_list)
    
    ## 交叉变异过程
    def ga_cross_next_group(self, ori_group, dict_score = 'ori', generation=0):
        new_dict = ori_group.copy()
        if generation == 0:
            score = 1.0 / len(ori_group)
            dict_score = {k:score for k in ori_group.keys()}
        g, p = np.array([[k, v] for k, v in dict_score.items()]).T
        cross_group = np.random.choice(g, size = int(len(g)/2), p = p, replace= False)
        for (fa,mo) in itertools.combinations(cross_group, 2):
            global GLOBAL_ORDER_ID
            GLOBAL_ORDER_ID += 1
            flag = GLOBAL_ORDER_ID
            fa_code, mo_code = ori_group[fa], ori_group[mo]
            cut_point = np.random.randint(1, len(fa_code)-1)
            fa_code0, fa_code1 = fa_code[:cut_point], fa_code[cut_point:]
            mo_code0, mo_code1 = mo_code[:cut_point], mo_code[cut_point:]
            new1 = np.hstack([fa_code0, mo_code1])
            prob = np.random.uniform(0, 1)
            if prob < self._change_prob:
                # 10%的变异
                count = int(len(new1) * self._cover_prob) + 1
                pos = np.random.randint(0, len(new1), count)
                #change_point = np.random.randint(0, len(new1))
                ## 改变该点的值
                ## new1[change_point] = not new1[change_point] 
                #所有的显性特征进行随机变异
                #new1 = [int(np.random.uniform(2, len(accumulators_pool) + 1)) if i> 0 else 0 for i in new1]
                new1[pos] = np.random.uniform(2, len(accumulators_pool) + 1, count)
            new_dict[flag] = np.array(new1)
        return new_dict
    
    
    # 基因突变计算
    def gevent_calc_factor(self, params):
        factor_data, factor_name, i = params
        if i > 0:
            accumulator = accumulators_pool[i](str(factor_name))
            sub_data = accumulator.transform(factor_data.set_index('trade_date'), 
                                                category_field='code', dropna=False)
            sub_data = sub_data['transformed']
        else:
            sub_data = factor_data[factor_name].copy()
        return sub_data.fillna(0), factor_name, i
     
    def gevent_evalue_group(self, params):
        sub_group, evalue_cols, total_data, diff_filed, g =  params
        index = 0
        res = {}
        cols = []
        jobs = []
        #此处引入gevent 
        for i in sub_group:
            if i > 0:
                factor_name = evalue_cols[index]
                factor_data = total_data[[str(factor_name),'code','trade_date']]
                jobs.append(gevent.spawn(self.gevent_calc_factor, 
                                         [factor_data, factor_name, i]))
            index += 1
        gevent.joinall(jobs)
        for result in jobs:
            factor_data, factor_name, i = result.value
            sub_factor_name = factor_name + 'c_' + str(i)
            if sub_factor_name not in res.keys():
                res[sub_factor_name] = factor_data.fillna(0).values
                cols.append(sub_factor_name)
        sub_data = pd.DataFrame(res)
        for diff in diff_filed:
            sub_data[diff] = total_data[diff]
        score = self._objective(sub_data.replace([np.inf, -np.inf], np.nan), cols)
        if score == np.nan: score = 0
        return score, cols, res, g
    
    def apply_calc_factor(self, data):
        data = data.copy()
        factor_name = data.name
        calc_acc = int(data['eval'])
        data.drop(['eval'],inplace=True)
        data = data.reset_index()
        if calc_acc > 0:
            accumulator = accumulators_pool[calc_acc]
            name = factor_name + 'c_' + str(calc_acc)
            sub_data = accumulator(factor_name).transform(data.fillna(0).set_index('trade_date'), name=name,
                                                          category_field='code',dropna=False)
            sub_data = sub_data.reset_index().set_index(['trade_date','code'])
        else:
            sub_data = data.set_index(['trade_date','code'])
        return sub_data
    
    #种群特征计算
    def process_evalue_group(self, params):
        sub_group, evalue_cols, total_data, diff_filed, g =  params
        cols = []
        #并行计算
        sub_factor_data = total_data.set_index(
                            ['trade_date','code'])[evalue_cols[[True if i> 0 else False for i in sub_group]]].T
        evalue_acc = [i for i in sub_group if i > 0]
        sub_factor_data['eval'] = evalue_acc
        start_time = time.time()
        result = sub_factor_data.apply(self.apply_calc_factor, axis=1)
        sub_data = pd.concat(list(result.values),axis=1).reset_index()
        MLog().write().info('%d: time:%f' %(int(g), time.time() - start_time))
        cols = [i for i in sub_data.columns if i not in ['trade_date','code']]
        diff_data = total_data[diff_filed]
        sub_data = sub_data.merge(diff_data, on=['code','trade_date'])
        #dataframe 转化为字典对应的数据
        score = self._objective(sub_data.replace([np.inf, -np.inf], np.nan), cols)
        if score == np.nan: score = 0
        return score, cols, sub_data, g
    
    ## 种群个体能力评价
    def ga_evalue_group(self, sub_group, total_data, evalue_cols, diff_filed):
        score_dict = {}
        cols_dict = {}
        jobs = []
        #multiprocessing
        start_time = time.time()
        for g, code in sub_group.items():
            jobs.append([sub_group[g], evalue_cols, total_data, 
                                     diff_filed,g])
        with multiprocessing.Pool(processes=self._parallel) as p:
            values_list = p.map(self.process_evalue_group, jobs)
        #values_list = []
        #for job in jobs:
        #    values_list.append(self.process_evalue_group(job))
        cols_sets = []
        factor_data = None
        print(time.time() - start_time)
        for values in values_list:
            score =  values[0]
            cols = values[1]
            sub_data = values[2]
            g = values[3]
            score_dict[g] = score
            cols_dict[g] = cols
            sub_data.sort_values(by=['trade_date','code'],ascending=True,inplace=True)
            if factor_data is None:
                cols_sets += cols
                factor_data = sub_data[cols].copy()
            else:
                diff_cols = list(set(cols) - set(cols_sets))
                cols_sets += diff_cols
                for diff in diff_cols:
                    factor_data[diff] = sub_data[diff].copy().values
        return factor_data, score_dict, cols_dict 
        
        
            
    
    #种群变异
    def ga_kill_group(self, total_data, ori_group, dict_score, evalue_cols, diff_filed, generation = 0):
        """
        total_data 种群数据集
        ori_group  有效种群特征组
        dict_score 分数
        evalue_cols 种群特征
        """
        # 生成新一代特征
        sub_group = self.ga_cross_next_group(ori_group, dict_score=dict_score,generation=generation)
        # 生成对应数据
        sub_data, ori_score_dict, cols_dict = self.ga_evalue_group(sub_group, total_data, evalue_cols, diff_filed)
        ori_score_se = pd.Series(ori_score_dict)
        MLog().write().info('Number of population:%d, Population average:%f'%(len(ori_score_se),ori_score_se.mean()))
        #真实分数
        fact_se = ori_score_se.sort_values(ascending=False)[:self._group_num] 
        #比例分数
        score_se = ori_score_se.sort_values(ascending=False)[:self._group_num] / (
            ori_score_se.sort_values(ascending= False)[:self._group_num].sum())
        
        liv_group = {}
        for k in score_se.keys():
            liv_group[k] = np.array([1 if i in cols_dict[k] else 0 for i in np.array(sub_data.columns)])
        
        factor_columns = sub_data.columns
        for diff in diff_filed:
            sub_data[diff] = total_data[diff].copy().values #pandas.core.series.Series拷贝时根据索引排序
        return sub_data, liv_group,  score_se,  dict(fact_se), np.array(factor_columns), cols_dict
        
    def genetic_run(self, total_data, diff_filed, sequence=['trade_date','code'],
                    factors_filed = None, strong_field = None, 
                    weak_field = None, is_best=True):
        if factors_filed is not None:
            point = int(np.random.uniform(0, len(factors_filed))/2)
            ori_field = factors_filed[:point]
            add_field = factors_filed[point:]
        elif strong_field is not None and weak_field is not None:
            ori_field = strong_field
            add_field = weak_field
        
        total_data = total_data.sort_values(by=sequence,ascending=True)
        evalue_cols = np.array(ori_field + add_field)
        last_score = 0.0
        #初始种群
        ori_group = {i:self.ga_generate_ori(ori_field, add_field) for i in range(self._group_num)}
        global GLOBAL_ORDER_ID
        GLOBAL_ORDER_ID += (len(ori_group) - 1)
        for i in range(self._generations):
            if i == 0:
                sub_data, liv_group, score_se, fact_se, cols, cols_dict = self.ga_kill_group(
                    total_data, ori_group, 'ori', evalue_cols, diff_filed, 0)
            else:
                sub_data, liv_group, score_se, fact_se, cols, cols_dict = self.ga_kill_group(
                     sub_data,   liv_group, score_se, cols, diff_filed, 1)
            
            best_fact_score = pd.Series(fact_se).sort_values(ascending=False)[0:1]
            best_score = pd.Series(score_se).sort_values(ascending=False)
            diff_fact_score = best_fact_score.values[0] - last_score
            diff_score = best_score[0:1].values[0] - best_score[-1:].values[0]
            now_best_score = (best_fact_score.values[0] if best_fact_score.values[0] > last_score else last_score)
            MLog().write().info('Multiplicaint number:%d, Best score:%f,%d(%f) And %d(%f) Best group difference:%f,%d best and wores difference:%f'%((i+1),
               now_best_score,
               i+1,
               best_fact_score.values[0], 
               i, last_score, 
               diff_fact_score,i+1,diff_score)) 
            if diff_fact_score < self._conver_prob and diff_score < self._conver_prob:
                break
            last_score = best_fact_score.values[0]
        
        if is_best:
            best_code = pd.Series(fact_se).sort_values(ascending=False)[0:1].index[0]
            best_list = liv_group[best_code]
            best_field = cols[[True if i> 0 else False for i in best_list]]
            return best_code, best_field
        else:
            field_group = {}
            for k, g in liv_group.items():
                field_group[k] = cols[[True if i> 0 else False for i in g]]
            return field_group



# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._Weekdays cimport Weekdays as ws

cpdef enum Weekdays:
    Sunday = ws.Sunday
    Monday = ws.Monday
    Tuesday = ws.Tuesday
    Wednesday = ws.Wednesday
    Thursday = ws.Thursday
    Friday = ws.Friday
    Saturday = ws.Saturday


# -*- coding: utf-8 -*-

import copy
from ultron.sentry.Analysis.SecurityValueHolders cimport SecurityStatelessSingleValueHolder
from ultron.sentry.Analysis.SecurityValueHolders cimport SecurityCombinedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders cimport build_holder
from ultron.sentry.Math.Accumulators.IAccumulators cimport Sign
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport XAverage
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport Average
from ultron.sentry.Math.Accumulators.StatefulAccumulators cimport MACD
from ultron.sentry.Math.Accumulators.IAccumulators cimport Exp
from ultron.sentry.Math.Accumulators.IAccumulators cimport Log
from ultron.sentry.Math.Accumulators.IAccumulators cimport Pow
from ultron.sentry.Math.Accumulators.IAccumulators cimport Sqrt
from ultron.sentry.Math.Accumulators.IAccumulators cimport Abs
from ultron.sentry.Math.Accumulators.IAccumulators cimport Acos
from ultron.sentry.Math.Accumulators.IAccumulators cimport Acosh
from ultron.sentry.Math.Accumulators.IAccumulators cimport Asin
from ultron.sentry.Math.Accumulators.IAccumulators cimport Asinh
from ultron.sentry.Math.Accumulators.IAccumulators cimport NormInv
from ultron.sentry.Math.Accumulators.IAccumulators cimport Ceil
from ultron.sentry.Math.Accumulators.IAccumulators cimport Floor
from ultron.sentry.Math.Accumulators.IAccumulators cimport Round
from ultron.sentry.Math.Accumulators.IAccumulators cimport Sigmoid
from ultron.sentry.Math.Accumulators.IAccumulators cimport Tanh
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport Diff
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport SimpleReturn
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport LogReturn
from ultron.sentry.Analysis.SeriesValues import s_maximum
from ultron.sentry.Analysis.SeriesValues import s_minimum


cdef class SecurityAverageValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityAverageValueHolder, self).__init__(holderType=Average,
                                                         x=x)


cdef class SecurityXAverageValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, window, x):
        super(SecurityXAverageValueHolder, self).__init__(holderType=XAverage,
                                                          x=x,
                                                          window=window)


cdef class SecurityMACDValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, short_win, long_win, x):
        super(SecurityMACDValueHolder, self).__init__(holderType=MACD,
                                                      x=x,
                                                      short_win=short_win,
                                                      long_win=long_win)


cdef class SecuritySignValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecuritySignValueHolder, self).__init__(holderType=Sign,
                                                      x=x)


cdef class SecurityExpValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x='x'):
        super(SecurityExpValueHolder, self).__init__(holderType=Exp,
                                                     x=x)


cdef class SecurityLogValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityLogValueHolder, self).__init__(holderType=Log,
                                                     x=x)


cdef class SecurityPowValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x, n=1):
        super(SecurityPowValueHolder, self).__init__(holderType=Pow,
                                                     x=x,
                                                     n=n)


cdef class SecuritySqrtValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecuritySqrtValueHolder, self).__init__(holderType=Sqrt,
                                                      x=x)


cdef class SecurityAbsValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityAbsValueHolder, self).__init__(holderType=Abs,
                                                     x=x)


cdef class SecurityAcosValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityAcosValueHolder, self).__init__(holderType=Acos,
                                                      x=x)


cdef class SecurityAcoshValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityAcoshValueHolder, self).__init__(holderType=Acosh,
                                                       x=x)


cdef class SecurityAsinValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityAsinValueHolder, self).__init__(holderType=Asin,
                                                      x=x)


cdef class SecurityAsinhValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityAsinhValueHolder, self).__init__(holderType=Asinh,
                                                       x=x)


cdef class SecurityNormInvValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x, fullAcc=False):
        super(SecurityNormInvValueHolder, self).__init__(holderType=NormInv,
                                                         x=x,
                                                         fullAcc=fullAcc)


cdef class SecurityCeilValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityCeilValueHolder, self).__init__(holderType=Ceil,
                                                      x=x)


cdef class SecurityFloorValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityFloorValueHolder, self).__init__(holderType=Floor,
                                                       x=x)


cdef class SecurityRoundValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityRoundValueHolder, self).__init__(holderType=Round,
                                                       x=x)


cdef class SecuritySigmoidValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecuritySigmoidValueHolder, self).__init__(holderType=Sigmoid,
                                                       x=x)
    
cdef class SecurityTanhValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityTanhValueHolder, self).__init__(holderType=Tanh,
                                                       x=x)
        
cdef class SecurityDiffValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityDiffValueHolder, self).__init__(holderType=Diff,
                                                      x=x)


cdef class SecuritySimpleReturnValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecuritySimpleReturnValueHolder, self).__init__(holderType=SimpleReturn,
                                                              x=x)


cdef class SecurityLogReturnValueHolder(SecurityStatelessSingleValueHolder):
    def __init__(self, x):
        super(SecurityLogReturnValueHolder, self).__init__(holderType=LogReturn,
                                                           x=x)


cdef class SecurityMaximumValueHolder(SecurityCombinedValueHolder):
    def __init__(self, x, y):
        super(SecurityMaximumValueHolder, self).__init__(x, y, s_maximum)

    def __str__(self):
        return "maximum({0}, {1})".format(str(self._left), str(self._right))


cdef class SecurityMinimumValueHolder(SecurityCombinedValueHolder):
    def __init__(self, x, y):
        super(SecurityMinimumValueHolder, self).__init__(x, y, s_minimum)

    def __str__(self):
        return "minimum({0}, {1})".format(str(self._left), str(self._right))



import os,time, pdb
import datetime as dt
import sqlalchemy
import numpy as np
import pandas as pd
import statsmodels.api as sm
from alphamind.data.processing import factor_processing
from alphamind.data.standardize import standardize
from alphamind.data.winsorize import winsorize_normal
from alphamind.data.neutralize import neutralize

class HighFrequencyWeighted(object):
    def __init__(self):
        self._industry_styles = ['Bank','RealEstate','Health','Transportation','Mining',
                                 'NonFerMetal','HouseApp','LeiService','MachiEquip','BuildDeco',
                                 'CommeTrade','CONMAT','Auto','Textile','FoodBever','Electronics',
                                 'Computer','LightIndus','Utilities','Telecom','AgriForest','CHEM',
                                 'Media','IronSteel','NonBankFinan','ELECEQP','AERODEF','Conglomerates']
        self._risk_styles = ['BETA','MOMENTUM','SIZE','EARNYILD','RESVOL','GROWTH','BTOP',
                             'LEVERAGE','LIQUIDTY','SIZENL']
    
    '''  
    def top_equal_weights(self, group, top='top', top_pct=0.2):
        if top == 'bottom':
            group = -1.0 * group
        group = group.rank(pct=True)
        group[group < 1.0-top_pct] = 0.0
        group[group >= 1.0-top_pct] = 1.0 / len(group[group >= 1.0-top_pct])
        return group
    
    def returns(self, factor_se, forward_returns, init_capital=100000):
        # 超额收益，此处benchmark采用全股票池算数均值
        forward_excess_returns = forward_returns.groupby(level=['trade_date']).apply(
            lambda x: x - x.mean())
        res_dict = {}
        for top in ['top', 'bottom']:
            weights = factor_se.groupby(level=['trade_date']).apply(lambda x: self.top_equal_weights(x, top=top, top_pct=0.2))
            weighted_returns = forward_excess_returns.multiply(weights, axis=0)
            factor_ret_se = weighted_returns.groupby(level='trade_date').sum()
            turnover_se = weights.unstack().diff().abs().sum(axis=1)
            res_dict[top] = pd.DataFrame({'returns': factor_ret_se, 'turnover': turnover_se})
        return res_dict
        
        
    def stats_information(self, price_res_dict, horizon):
        stats_df = pd.DataFrame({x: self.calc_stats(price_res_dict[x], horizon) for x in price_res_dict.keys()}).T
        stats_df.index.name = 'top_bot'
        #stats_df = stats_df.reset_index()
        return stats_df
    '''
    
    def to_weights(self, group):
        demeaned_vals = group - group.mean()
        return demeaned_vals / demeaned_vals.abs().sum()
    
    def to_ls_count(self, group, long=True):
        demeaned_vals = group - group.mean()
        if long:
            count = len(demeaned_vals[demeaned_vals>0])
        else:
            count = len(demeaned_vals[demeaned_vals<0])
        return count 
    
    def returns(self, factor_se, forward_returns):
        weights = factor_se.groupby(level=['trade_date']).apply(self.to_weights)
        weighted_returns = forward_returns.multiply(weights, axis=0)
        
        factor_ret_se = weighted_returns.groupby(level='trade_date').sum()
        turnover_se = weights.unstack().diff().abs().sum(axis=1)
        long_count = factor_se.groupby(level=['trade_date']).apply(self.to_ls_count)
        short_count = factor_se.groupby(level=['trade_date']).apply(lambda x: self.to_ls_count(x, long=False))
        #计算方向
        forward_returns.name='nxt1_ret'
        factor_se.name = 'factor'
        forward_returns.reset_index().merge(factor_se.reset_index(),on=['trade_date','code'])
        ic_mean = forward_returns.reset_index().merge(factor_se.reset_index(),
                                            on=['trade_date','code']).groupby('trade_date').apply(
            lambda x: np.corrcoef(x['factor'], x['nxt1_ret'])[0, 1]).mean()
        direction = np.sign(ic_mean)
        return factor_ret_se, turnover_se, long_count, short_count, direction
    
    def calc_stats(self, returns_df, turnover_df, direction, horizon):
        # 总体指标
        returns_se, turnover_se = returns_df, turnover_df
        #纯多头组合可以考虑费率，多空组合不能考虑费率，若为负收益率(反方向因子),再减费率不符合逻辑
        #returns_se = returns_se - turnover_se * 0.001
        
        ir = returns_se.mean() / returns_se.std()
        sharpe = np.sqrt(252 / horizon) * ir
        turnover = turnover_se.mean()
        returns = returns_se.sum() * 252 / horizon / len(returns_se)
        fitness = sharpe * np.sqrt(abs(returns) / turnover)
        margin = returns_se.sum() / turnover_se.sum()
        stats_se = pd.Series({'status':1, 'ir': ir, 'sharpe': sharpe, 'turnover': turnover, 
                              'returns': returns, 'fitness': fitness, 'margin': margin,
                              'direction':direction,
                              'score':abs(sharpe)})
        return stats_se
        
    def run(self, factor_data, risk_data, mkt_df, default_value,
            factor_name, is_neutralize = True, risk_styles = None, horizon=1, 
            keys='bar30_vwap', method='quantile', up_limit=0.025, down_limit=0.025,
            init_capital=100000):
        """
        参数：
            horizon: 调仓期，按照交易日计算。
        """
        fac_list = [factor_name]
        factors_se = factor_data.merge(risk_data.reset_index(), on=['code','trade_date'])
        ndiff_field =  [i for i in list(set(factors_se.columns)) if i not in fac_list]
        risk_styles = self._risk_styles if risk_styles is None else risk_styles
        alpha_res = []
        grouped = factors_se.groupby(['trade_date'])
        for k, g in grouped:
            if is_neutralize:
                ret_preprocess = neutralize(g[self._industry_styles + risk_styles].values.astype(float),
                                factor_processing(g[fac_list].fillna(0).values,
                                                  pre_process=[winsorize_normal, standardize]))
            else:
                ret_preprocess = factor_processing(g[fac_list].fillna(0).values,
                                                  pre_process=[winsorize_normal, standardize])
            f = pd.DataFrame(ret_preprocess, columns=fac_list)
            for k in ndiff_field:
                f[k] = g[k].values
            alpha_res.append(f)
        factor_se = pd.concat(alpha_res).set_index(['trade_date','code'])[factor_name]
        mkt_se = mkt_df.set_index(['trade_date','code'])
        price_tb = mkt_se[keys].unstack()
        return_tb = (price_tb.shift(-horizon) / price_tb - 1.0)
        return_tb[return_tb>10.0] = np.NaN
        return_tb = return_tb.shift(-1)
        return_se = return_tb.stack()
        #  多空计算因子收益数据
        factor_ret_se, turnover_se, long_count, short_count, direction = self.returns(
            factor_se, return_se)
        return self.calc_stats(factor_ret_se, turnover_se, direction, horizon)
        # top 20% 等权方法计算的因子收益数据
        '''
        price_res_dict = self.returns(factor_se, return_se)
        stats_df = self.stats_information(price_res_dict, horizon)
        if stats_df.loc['top'].score > stats_df.loc['bottom'].score:# and stats_df.loc['top'].score > 0:
            return stats_df.loc['top'].to_dict()
        elif stats_df.loc['bottom'].score > stats_df.loc['top'].score:# and stats_df.loc['bottom'].score > 0:
            return stats_df.loc['bottom'].to_dict()
        else:
            return {'status':0, 'score':default_value}
        '''



# -*- coding: utf-8 -*-


from enum import IntEnum
from math import inf
from typing import Dict
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
import pandas as pd
from deprecated import deprecated

def require(cond: bool, exception=AssertionError, msg: str = None):
    if not cond:
        raise exception(msg)


class BoundaryDirection(IntEnum):
    LOWER = -1
    UPPER = 1


class BoundaryType(IntEnum):
    ABSOLUTE = 0
    RELATIVE = 1
    MAXABSREL = 2
    MINABSREL = 3

    
class BoundaryImpl(object):

    def __init__(self,
                 direction: BoundaryDirection,
                 b_type: BoundaryType,
                 val):
        self.direction = direction
        self.b_type = b_type
        self.val = val
        self._validation()

    def _validation(self):
        require(
            self.b_type in [BoundaryType.ABSOLUTE, BoundaryType.RELATIVE, BoundaryType.MAXABSREL,
                            BoundaryType.MINABSREL],
            ValueError,
            "Boundary Type {0} is not recognized".format(self.b_type))

        require(
            self.direction == BoundaryDirection.LOWER or self.direction == BoundaryDirection.UPPER,
            ValueError,
            "Boundary direction {0} is not recognized".format(self.direction))

    def __call__(self, center: float):
        if self.b_type == BoundaryType.ABSOLUTE:
            return self.val + center
        elif self.b_type == BoundaryType.MAXABSREL:
            abs_threshold = self.val[0]
            rel_threshold = self.val[1]
            if self.direction == BoundaryDirection.LOWER:
                rel_bound = center - abs(center) * rel_threshold
                abs_bound = center - abs_threshold
                return min(rel_bound, abs_bound)
            elif self.direction == BoundaryDirection.UPPER:
                rel_bound = center + abs(center) * rel_threshold
                abs_bound = center + abs_threshold
                return max(rel_bound, abs_bound)
        elif self.b_type == BoundaryType.MINABSREL:
            abs_threshold = self.val[0]
            rel_threshold = self.val[1]
            if self.direction == BoundaryDirection.LOWER:
                rel_bound = center - abs(center) * rel_threshold
                abs_bound = center - abs_threshold
                return max(rel_bound, abs_bound)
            elif self.direction == BoundaryDirection.UPPER:
                rel_bound = center + abs(center) * rel_threshold
                abs_bound = center + abs_threshold
                return min(rel_bound, abs_bound)
        else:
            require(center >= 0., ValueError,
                    "relative bounds only support positive back bone value")
            return self.val * center


class BoxBoundary(object):

    def __init__(self,
                 lower_bound: BoundaryImpl,
                 upper_bound: BoundaryImpl):
        self.lower = lower_bound
        self.upper = upper_bound

    def bounds(self, center):
        l_b, u_b = self.lower(center), self.upper(center)
        require(l_b <= u_b, ValueError, "lower bound should be lower then upper bound")
        return l_b, u_b


def create_box_bounds(names: List[str],
                      b_type: Union[Iterable[BoundaryType], BoundaryType],
                      l_val: Union[Iterable[float], float],
                      u_val: Union[Iterable[float], float]) -> Dict[str, BoxBoundary]:
    """
    helper function to quickly create a series of bounds
    """
    bounds = dict()

    if not hasattr(b_type, '__iter__'):
        b_type = np.array([b_type] * len(names))

    if not hasattr(l_val, '__iter__'):
        l_val = np.array([l_val] * len(names))

    if not hasattr(u_val, '__iter__'):
        u_val = np.array([u_val] * len(names))

    for i, name in enumerate(names):
        lower = BoundaryImpl(BoundaryDirection.LOWER,
                             b_type[i],
                             l_val[i])
        upper = BoundaryImpl(BoundaryDirection.UPPER,
                             b_type[i],
                             u_val[i])
        bounds[name] = BoxBoundary(lower, upper)
    return bounds


class LinearConstraints(object):

    def __init__(self,
                 bounds: Dict[str, BoxBoundary],
                 cons_mat: pd.DataFrame,
                 backbone: np.ndarray = None):
        self.names = list(set(bounds.keys()).intersection(set(cons_mat.columns)))
        self.bounds = bounds
        self.cons_mat = cons_mat
        self.backbone = backbone

        require(cons_mat.shape[0] == len(backbone) if backbone is not None else True,
                "length of back bond should be same as number of rows of cons_mat")

    def risk_targets(self) -> Tuple[np.ndarray, np.ndarray]:
        lower_bounds = []
        upper_bounds = []

        if self.backbone is None:
            backbone = np.zeros(len(self.cons_mat))
        else:
            backbone = self.backbone

        for name in self.names:
            center = backbone @ self.cons_mat[name].values
            l, u = self.bounds[name].bounds(center)
            lower_bounds.append(l)
            upper_bounds.append(u)
        return np.array(lower_bounds), np.array(upper_bounds)

    @property
    def risk_exp(self) -> np.ndarray:
        return self.cons_mat[self.names].values


@deprecated(
    reason="Constraints is deprecated in alpha-mind 0.1.1. Please use LinearConstraints instead.")
class Constraints(object):

    def __init__(self,
                 risk_exp: Optional[np.ndarray] = None,
                 risk_names: Optional[np.ndarray] = None):
        self.risk_exp = risk_exp

        if risk_names is not None:
            self.risk_names = np.array(risk_names)
        else:
            self.risk_names = np.array([])

        n = len(self.risk_names)

        self.risk_maps = dict(zip(self.risk_names, range(n)))
        self.lower_bounds = -inf * np.ones(n)
        self.upper_bounds = inf * np.ones(n)

    def set_constraints(self, tag: str, lower_bound: float, upper_bound: float):
        index = self.risk_maps[tag]
        self.lower_bounds[index] = lower_bound
        self.upper_bounds[index] = upper_bound

    def add_exposure(self, tags: np.ndarray, new_exp: np.ndarray):
        if len(tags) != new_exp.shape[1]:
            raise ValueError(
                'new dags length is not compatible with exposure shape {1}'.format(len(tags),
                                                                                   new_exp.shape))

        for tag in tags:
            if tag in self.risk_maps:
                raise ValueError('tag {0} is already in risk table'.format(tag))

        self.risk_names = np.concatenate((self.risk_names, tags))

        if self.risk_exp is not None:
            self.risk_exp = np.concatenate((self.risk_exp, new_exp), axis=1)
        else:
            self.risk_exp = new_exp

        n = len(self.risk_names)
        self.risk_maps = dict(zip(self.risk_names, range(n)))

        self.lower_bounds = np.concatenate((self.lower_bounds, -inf * np.ones(len(tags))))
        self.upper_bounds = np.concatenate((self.upper_bounds, inf * np.ones(len(tags))))

    def risk_targets(self) -> Tuple[np.ndarray, np.ndarray]:
        return self.lower_bounds, self.upper_bounds


# -*- coding: utf-8 -*-

class BaseModule(object):
    def __init__(self, name, wid, token, redis_client):
        __str__ = name
        self._redis_client = redis_client
        self._wid = wid
        self._token = token


# -*- coding: utf-8 -*-
import six
import base64
import zlib
import time
import datetime
import os
import json
import pdb
from ultron.utilities.redis.redis_client import RedisClient
from ultron.utilities.singleton import Singleton
from ultron.utilities.zlib_engine import zip_compress,unzip_compress
from ultron.utilities.short_uuid import uuid, decode
from ultron.config import config_setting

@six.add_metaclass(Singleton)
class SubmitTask(object):
    def __init__(self, **kwargs):
        if ('host' in kwargs) and ('port' in kwargs) and ('pwd' in kwargs):
            self._redis_client = RedisClient(host=kwargs['host'],
                                          port=kwargs['port'],
                                          password=kwargs['pwd'])
        else:
            self._redis_client = RedisClient(host=config_setting.queue_host,
                                          port=config_setting.queue_port,
                                          password=config_setting.queue_pwd)
    
    def submit_packet(self, uid, packet):
        upload_info = {}
        upload_info['name'] = 'packet'
        upload_info['opcode'] = 'upload_packet'
        upload_info['uid'] = uid
        upload_info['packet_name'] = packet
        file_name = str(decode(uuid())) + '.uldk'
        upload_info['file_name'] = file_name
        task_id = uuid()
        # 压缩
        zip_compress(packet, file_name)
        #上传文件
        with open(os.path.join('./', file_name) ,'rb') as f:
            content = f.read()
            upload_info['file_info'] = base64.b64encode(content).decode()
        #删除打包文件
        if os.path.isfile(file_name):
            os.remove(file_name)
            
        return self._redis_client.hset('ultron:work:update',str(task_id),json.dumps(upload_info))
        
    
submit_task = SubmitTask()


# -*- coding: utf-8 -*-
import pdb
import datetime
import json
import hashlib
from ultron.cluster.central.extern_modules.base_module import BaseModule
from ultron.cluster.central.extern_modules.middleware.module import ModelsSingleton
class Module(BaseModule):
    def __init__(self, name, redis_client):
        self._secret_key = 'd6f89b09'
        super(Module, self).__init__(name, redis_client)
        self._func = {'login_in':self.login_in,
                     'heart_tick':self.heart_tick}
        
    
    def process_respone(self, result):
        name = result['name']
        opcode = result['opcode']
        self._func[opcode](result)
    
    def login_in(self, result):
        wid = result['wid']
        login_info = {'name':'login','opcode':'login_info','result':'sucess'}
        queue_name = 'ultron:work:work_id:' + wid
        work_info = {'wid':wid,'token':result['token'],
                     'queue_name':queue_name,
                    'login_time':result['login_time'],
                    'update_time':datetime.datetime.now()}
        print(wid,result,queue_name)
        #节点登录认证 比较两个self._token, token是否相等
        token = hashlib.sha1((self._secret_key + wid.replace('-','')).encode()).hexdigest()
        
        ModelsSingleton().add_cluster(wid, work_info)
        self._redis_client.hset(queue_name, wid, json.dumps(login_info))
     
    def heart_tick(self, result):
        ModelsSingleton().set_alive(result['wid'])


# -*- coding: utf-8 -*-

from ultron.tradingday import Date


class SettingsFactory(object):
    def __init__(self):
        self._evaluationDate = None
        self._includeToday = None

    @property
    def evaluationDate(self):
        if not self._evaluationDate:
            return Date.todaysDate()
        return self._evaluationDate

    @evaluationDate.setter
    def evaluationDate(self, value):
        #require(isinstance(value, Date), ValueError, "{0} is not a valid PyFin date object".format(value))
        self._evaluationDate = value

    @property
    def includeTodaysCashFlows(self):
        return self._includeToday

    @includeTodaysCashFlows.setter
    def includeTodaysCashFlows(self, value):
        self._evaluationDate = value

    def resetEvaluationDate(self):
        self._evaluationDate = None

    def anchorEvaluationDate(self):
        if self._evaluationDate is None:
            self._evaluationDate = Date.todaysDate()


Settings = SettingsFactory()



# -*- coding: utf-8 -*-

from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySignValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityXAverageValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMACDValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityExpValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySqrtValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityPowValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAbsValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcosValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAcoshValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAsinhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityNormInvValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityCeilValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityFloorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityDiffValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySigmoidValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityTanhValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySimpleReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogReturnValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMaximumValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityMinimumValueHolder

from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingDecay
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMax
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingArgMin
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRank
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingQuantile
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAllTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingAnyTrue
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingSum
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingVariance
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingStandardDeviation
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCountedPositive
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCountedNegative
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingNegativeAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingPositiveDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingNegativeDifferenceAverage
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingRSI
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingLogReturn
from ultron.sentry.Analysis.TechnicalAnalysis.StatefulTechnicalAnalysers import SecurityMovingCorrelation

__all__ = ['SecuritySignValueHolder',
           'SecurityAverageValueHolder',
           'SecurityXAverageValueHolder',
           'SecurityMACDValueHolder',
           'SecurityExpValueHolder',
           'SecurityLogValueHolder',
           'SecuritySqrtValueHolder',
           'SecurityPowValueHolder',
           'SecurityAbsValueHolder',
           'SecurityAcosValueHolder',
           'SecurityAcoshValueHolder',
           'SecurityAsinValueHolder',
           'SecurityAsinhValueHolder',
           'SecurityNormInvValueHolder',
           'SecurityCeilValueHolder',
           'SecurityFloorValueHolder',
           'SecurityRoundValueHolder',
           'SecurityDiffValueHolder',
           'SecurityTanhValueHolder',
           'SecuritySigmoidValueHolder',
           'SecuritySimpleReturnValueHolder',
           'SecurityLogReturnValueHolder',
           'SecurityMaximumValueHolder',
           'SecurityMinimumValueHolder',
           'SecurityMovingAverage',
           'SecurityMovingDecay',
           'SecurityMovingMax',
           'SecurityMovingArgMax',
           'SecurityMovingMin',
           'SecurityMovingArgMin',
           'SecurityMovingRank',
           'SecurityMovingQuantile',
           'SecurityMovingAllTrue',
           'SecurityMovingAnyTrue',
           'SecurityMovingSum',
           'SecurityMovingVariance',
           'SecurityMovingStandardDeviation',
           'SecurityMovingCountedPositive',
           'SecurityMovingPositiveAverage',
           'SecurityMovingCountedNegative',
           'SecurityMovingNegativeAverage',
           'SecurityMovingPositiveDifferenceAverage',
           'SecurityMovingNegativeDifferenceAverage',
           'SecurityMovingRSI',
           'SecurityMovingLogReturn',
           'SecurityMovingCorrelation']


# -*- coding: utf-8 -*-

from ultron.sentry.Math import Accumulators
from ultron.sentry.Math import Distributions
from ultron.sentry.Math import MathConstants

__all__ = ['Accumulators',
           'Distributions',
           'MathConstants']


# -*- coding: utf-8 -*-

import numpy as np
import pandas as pd


def simple_settle(weights: np.ndarray,
                  dx_return: np.ndarray,
                  groups: np.ndarray=None,
                  benchmark: np.ndarray=None) -> pd.DataFrame:

    weights = weights.flatten()
    dx_return = dx_return.flatten()

    if benchmark is not None:
        net_pos = weights - benchmark
    else:
        net_pos = weights

    ret_arr = net_pos * dx_return

    if groups is not None:
        ret_agg = pd.Series(ret_arr).groupby(groups).sum()
        ret_agg.loc['total'] = ret_agg.sum()
    else:
        ret_agg = pd.Series(ret_arr.sum(), index=['total'])

    ret_agg.index.name = 'industry'
    ret_agg.name = 'er'

    pos_table = pd.DataFrame(net_pos, columns=['weight'])
    pos_table['ret'] = dx_return

    if groups is not None:
        ic_table = pos_table.groupby(groups).corr()['ret'].loc[(slice(None), 'weight')]
        ic_table.loc['total'] = pos_table.corr().iloc[0, 1]
    else:
        ic_table = pd.Series(pos_table.corr().iloc[0, 1], index=['total'])

    return pd.DataFrame({'er': ret_agg.values,
                         'ic': ic_table.values},
                        index=ret_agg.index)


# -*- coding: utf-8 -*-

from ultron.sentry.Analysis import transform
from ultron.sentry.Analysis.SeriesValues import SeriesValues
from ultron.sentry.api.Analysis import SIGN
from ultron.sentry.api.Analysis import AVG
from ultron.sentry.api.Analysis import EMA
from ultron.sentry.api.Analysis import MACD
from ultron.sentry.api.Analysis import RSI
from ultron.sentry.api.Analysis import MCORR
from ultron.sentry.api.Analysis import MA
from ultron.sentry.api.Analysis import MADecay
from ultron.sentry.api.Analysis import MMAX
from ultron.sentry.api.Analysis import MARGMAX
from ultron.sentry.api.Analysis import MMIN
from ultron.sentry.api.Analysis import MARGMIN
from ultron.sentry.api.Analysis import MRANK
from ultron.sentry.api.Analysis import MAXIMUM
from ultron.sentry.api.Analysis import MINIMUM
from ultron.sentry.api.Analysis import MQUANTILE
from ultron.sentry.api.Analysis import MALLTRUE
from ultron.sentry.api.Analysis import MANYTRUE
from ultron.sentry.api.Analysis import MSUM
from ultron.sentry.api.Analysis import MVARIANCE
from ultron.sentry.api.Analysis import MSTD
from ultron.sentry.api.Analysis import MNPOSITIVE
from ultron.sentry.api.Analysis import MAPOSITIVE
from ultron.sentry.api.Analysis import CURRENT
from ultron.sentry.api.Analysis import LAST
from ultron.sentry.api.Analysis import SQRT
from ultron.sentry.api.Analysis import DIFF
from ultron.sentry.api.Analysis import RETURNSimple
from ultron.sentry.api.Analysis import RETURNLog
from ultron.sentry.api.Analysis import EXP
from ultron.sentry.api.Analysis import LOG
from ultron.sentry.api.Analysis import POW
from ultron.sentry.api.Analysis import ABS
from ultron.sentry.api.Analysis import ACOS
from ultron.sentry.api.Analysis import ACOSH
from ultron.sentry.api.Analysis import ASIN
from ultron.sentry.api.Analysis import ASINH
from ultron.sentry.api.Analysis import NORMINV
from ultron.sentry.api.Analysis import CEIL
from ultron.sentry.api.Analysis import FLOOR
from ultron.sentry.api.Analysis import ROUND
from ultron.sentry.api.Analysis import SHIFT
from ultron.sentry.api.Analysis import DELTA
from ultron.sentry.api.Analysis import IIF
from ultron.sentry.api.Analysis import INDUSTRY

from ultron.sentry.api.Analysis import CSRank
from ultron.sentry.api.Analysis import CSTopN
from ultron.sentry.api.Analysis import CSBottomN
from ultron.sentry.api.Analysis import CSTopNQuantile
from ultron.sentry.api.Analysis import CSBottomNQuantile
from ultron.sentry.api.Analysis import CSMean
from ultron.sentry.api.Analysis import CSMeanAdjusted
from ultron.sentry.api.Analysis import CSQuantiles
from ultron.sentry.api.Analysis import CSZScore
from ultron.sentry.api.Analysis import CSFillNA
from ultron.sentry.api.Analysis import CSRes

from ultron.sentry.Utilities.Asserts import pyFinAssert


__all__ = ["transform",
           "SIGN",
           "SeriesValues",
           "AVG",
           "EMA",
           "MACD",
           "RSI",
           "MCORR",
           "MA",
           "MADecay",
           "MMAX",
           "MARGMAX",
           "MMIN",
           "MARGMIN",
           "MRANK",
           "MAXIMUM",
           "MINIMUM",
           "MQUANTILE",
           "MALLTRUE",
           "MANYTRUE",
           "MSUM",
           "MVARIANCE",
           "MSTD",
           "MNPOSITIVE",
           "MAPOSITIVE",
           "CURRENT",
           "LAST",
           "SQRT",
           "DIFF",
           "RETURNSimple",
           "RETURNLog",
           "EXP",
           "LOG",
           "POW",
           "ABS",
           "ACOS",
           "ACOSH",
           "ASIN",
           "ASINH",
           "NORMINV",
           "CEIL",
           "FLOOR",
           "ROUND",
           "SHIFT",
           "DELTA",
           "IIF",
           "INDUSTRY",
           "CSRank",
           "CSTopN",
           "CSBottomN",
           "CSTopNQuantile",
           "CSBottomNQuantile",
           "CSMean",
           "CSMeanAdjusted",
           "CSQuantiles",
           "CSZScore",
           "CSFillNA",
           "CSRes",
           "pyFinAssert"]



# -*- coding: utf-8 -*-
import pandas as pd
def calc_ic(factor_df, return_df, factor_list, return_col_name='target_return', ic_type='spearman'):
    """
    计算因子IC值, 本月和下月因子值的秩相关
    params:
            factor_df: DataFrame, columns=['ticker', 'tradeDate', factor_list]
            return_df: DataFrame, colunms=['ticker, 'tradeDate'， return_col_name], 预先计算好的未来的收益率
            factor_list:　list， 需要计算IC的因子名list
            return_col_name: str, return_df中的收益率列名
            method: : {'spearman', 'pearson'}, 默认'spearman', 指定计算rank IC('spearman')或者Normal IC('pearson')
    return:
            DataFrame, 返回各因子的IC序列， 列为: ['tradeDate', factor_list]
    """
    merge_df = factor_df.merge(return_df, on=['code', 'trade_date'])
    # 遍历每个因子，计算对应的IC
    factor_ic_list = []
    for factor_name in factor_list:
        tmp_factor_ic = merge_df.groupby(['trade_date']).apply(
            lambda x: x[[factor_name, return_col_name]].corr(method=ic_type).values[0, 1])
        tmp_factor_ic.name = factor_name
        factor_ic_list.append(tmp_factor_ic)
    factor_ic_frame = pd.concat(factor_ic_list, axis=1)
    factor_ic_frame.reset_index(inplace=True)
    return factor_ic_frame



import numpy as np
import pandas as pd
from ultron.factor.data.processing import factor_processing
from ultron.factor.data.standardize import standardize
from ultron.factor.data.winsorize import winsorize_normal
from ultron.factor.data.neutralize import neutralize
from ultron.factor.combine.kutil import calc_ic
import pdb


class IC_Weighted(object):
    def __init__(self, result_type):
        self._result_type = result_type
        self._industry_styles = ['Bank','RealEstate','Health','Transportation','Mining',
                                 'NonFerMetal','HouseApp','LeiService','MachiEquip','BuildDeco',
                                 'CommeTrade','CONMAT','Auto','Textile','FoodBever','Electronics',
                                 'Computer','LightIndus','Utilities','Telecom','AgriForest','CHEM',
                                 'Media','IronSteel','NonBankFinan','ELECEQP','AERODEF','Conglomerates']
        self._risk_styles = ['BETA','MOMENTUM','SIZE','EARNYILD','RESVOL','GROWTH','BTOP',
                             'LEVERAGE','LIQUIDTY','SIZENL']
        
        
    #暂时写此处，去极值
    def winsorize(self, total_data, method='sigma', limits=(3.0, 3.0), drop=True):
        se = total_data.copy()
        if method == 'quantile':
            down, up = se.quantile([limits[0], 1.0 - limits[1]])
        elif method == 'sigma':
            std, mean = se.std(), se.mean()
            down, up = mean - limits[0]*std, mean + limits[1]*std
        
        if drop:
            se[se<down] = np.NaN
            se[se>up] = np.NaN
        else:
            se[se<down] = down
            se[se>up] = up
        return se
    
    #中性化
    def neutralize(self, se, risk_df):
        se_index = se.index
        se = se.dropna()
        risk = risk_df.loc[se.index]

        # use numpy for neu, which is faster
        x = np.linalg.lstsq(risk.values, np.matrix(se).T, rcond=None)[0]
        se_neu = se - risk.dot(x)[0]
    
        # # use statsmodels for neu
        # x = risk.values
        # y = se.values
        # model = sm.OLS(y, x)
        # results = model.fit()
        # se_neu = pd.Series(data=y - x.dot(results.params), index=se.index)
        
        return se_neu.reindex(se_index)
    
    #标准化
    def standardize(self, total_data):
        try:
            res = (total_data - np.nanmean(total_data)) / np.nanstd(total_data)
        except:
            res = pd.Series(data=np.NaN, index=total_data.index)
        return res
    
    
    def ic_calc(self, factor_data, forward_returns, risk_data, factor_name,
            risk_styles = None, method='quantile', up_limit=0.025, down_limit=0.025,
            return_col_name='nxt1_ret',ic_type='spearman'):
        fac_list = [factor_name]
        factors_se = factor_data.merge(risk_data.reset_index(), on=['code','trade_date'])
        ndiff_field =  [i for i in list(set(factors_se.columns)) if i not in fac_list]
        risk_styles = self._risk_styles if risk_styles is None else risk_styles
        alpha_res = []
        grouped = factors_se.groupby(['trade_date'])
        for k, g in grouped:
            ret_preprocess = neutralize(g[self._industry_styles + risk_styles].values.astype(float),
                                factor_processing(g[fac_list].fillna(0).values,
                                                  pre_process=[winsorize_normal, standardize]))
            f = pd.DataFrame(ret_preprocess, columns=fac_list)
            for k in ndiff_field:
                f[k] = g[k].values
            alpha_res.append(f)
        factor_se = pd.concat(alpha_res).set_index(['trade_date','code'])[fac_list]
        forward_returns = forward_returns.set_index(['trade_date','code'])
        return calc_ic(factor_df=factor_se.reset_index(), return_df=forward_returns, 
            factor_list=fac_list, return_col_name=return_col_name, ic_type=ic_type)


    def run(self, factor_data, forward_returns, risk_data, factor_name, is_neutralize=True,
            risk_styles = None, method='quantile', up_limit=0.025, down_limit=0.025,
            return_col_name='nxt1_ret',ic_type='spearman'):
        ic_serliaze = self.ic_calc(factor_data=factor_data,forward_returns=forward_returns,
                                  risk_data=risk_data,factor_name=factor_name,
                                  risk_styles=risk_styles,method=method,up_limit=up_limit,
                                  down_limit=down_limit,return_col_name=return_col_name,
                                  ic_type=ic_type)[factor_name].dropna()
        return ic_serliaze.values.mean() if self._result_type=='ic' else ic_serliaze.values.mean() / ic_serliaze.values.std()



# -*- coding: utf-8 -*-

import threading,logging,datetime,os, six
from ultron.utilities.singleton import Singleton
@six.add_metaclass(Singleton)
class MLog(object):
    
    def __init__(self):
        self._logging = None #self._config(level=logging.DEBUG)
    
    def config(self, name="logging", level=logging.DEBUG):
        self._logging = self._config(name=name, level=level)

    def _config(self, name="logging", level=logging.DEBUG):
        """
        Constructor
        """
        dir_name = os.path.expandvars('$HOME') + '/MLOG/' + name + '/'
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        filename = dir_name + name + '_' + datetime.datetime.now().strftime('%b_%d_%H_%M')+'.log'
        #format_str = "[%(process)d %(thread)d][%(asctime)s][%(filename)s line:%(lineno)d][%(levelname)s] %(message)s"
        format_str = "[%(filename)s line:%(lineno)d][%(levelname)s] %(message)s"
        # define a Handler which writes INFO messages or higher to the sys.stderr
        logging.basicConfig(level=level,
                            format=format_str,
                            datefmt='%m-%d %H:%M',
                            filename=filename,
                            filemode='w')
        console = logging.StreamHandler()
        console.setLevel(level)
        formatter = logging.Formatter(format_str)
        console.setFormatter(formatter)
        # 将定义好的console日志handler添加到root logger
        logging.getLogger('').addHandler(console)
        return logging

    def write(self):
        return self._logging



# -*- coding: utf-8 -*-

# 列名: 股票代码 code, 日期 trade_date
import pandas as pd
import numpy as np
import seaborn as sns
import scipy.stats as st
import statsmodels.api as sm 
import seaborn, pdb
from sklearn.covariance import LedoitWolf
from cvxopt import matrix, solvers
from . kutil import calc_ic

def factor_combine(factor_df, factor_list, weight_df):
    """
    因子合成
    参数:
        factor_df: DataFrame, 待合成因子值
        factor_list: list, 待合成因子列表
        weight_df:　DataFrame, 因子权重
    返回:
        DataFrame, 复合因子
    """
    factor_df = factor_df.dropna(subset=factor_list, how='all')
    
    merge_df = factor_df.merge(weight_df, on=['trade_date'], suffixes=('', '_w'))
    
    w_list = [fn+'_w' for fn in factor_list]
    merge_df[w_list] = np.where(merge_df[factor_list].isnull(), 0, merge_df[w_list])
    merge_df['combine'] = np.sum(merge_df[factor_list].fillna(0).values*merge_df[w_list].values, axis=1)
    merge_df['combine'] = merge_df['combine']/merge_df[w_list].abs().sum(axis=1)
    
    return merge_df[['code', 'trade_date', 'combine']]

def equal_combine(factor_df, factor_list, ):
    """
    等权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        factor_list: list, 待合成因子列表
    返回:
        DataFrame, 复合因子
    """
    factor_df = factor_df.copy()
    return factor_df[factor_list].mean(axis=1).values
    #factor_df['equal']= factor_df[factor_list].mean(axis=1)
    #return factor_df[['code', 'trade_date', 'equal']]


def hist_ret_combine(factor_df, mret_df, factor_list, span,
                     size_indu_df=None, 
                     indu_list=[],
                     method='equal', half_life=1):
    """
    历史收益率加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算历史收益率均值
        method: 历史收益率均值计算方法。'equal':算术平均；'half_life':半衰加权
        half_life: int, 半衰加权时的半衰期
    返回:
        DataFrame, 复合因子
    """
    #merged_df = factor_df.merge(size_indu_df, on=['code', 'trade_date']).merge(mret_df, on=['code', 'trade_date'])
    merged_df = factor_df.merge(mret_df, on=['code', 'trade_date'])
    merged_df = merged_df.dropna(subset=['nxt1_ret'])
    # 计算历史因子收益率
    hist_ret = []
    for fn in factor_list:
        tmp = merged_df.groupby('trade_date').apply(lambda df: sm.OLS(df['nxt1_ret'], sm.add_constant(df[[fn]+indu_list]), missing='drop').fit().params[1])
        hist_ret.append(tmp)
    hist_ret = pd.concat(hist_ret, axis=1)
    hist_ret.columns = factor_list
    hist_ret = hist_ret.reset_index().sort_values('trade_date')
    hist_ret['trade_date'] = hist_ret['trade_date'].shift(-1)
    
    # 计算历史因子收益率移动平均
    for fn in factor_list:
        if method == 'equal':
            hist_ret[fn+'_ma'] = hist_ret[fn].rolling(span).mean()
        elif method == 'half_life':
            hist_ret[fn+'_ma'] = hist_ret[fn].rolling(span).apply(lambda x: np.average(x, weights=list(reversed([0.5**(1.0*i/half_life) for i in range(span)]))))
        else:
            return 0
    
    hist_ret = hist_ret.dropna()
    hist_ret = hist_ret.drop(factor_list, axis=1)
    hist_ret.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, hist_ret)
    return conb_df, hist_ret


def hist_ic_combine(factor_df, mret_df, factor_list, span, method='equal', half_life=1):
    """
    历史收IC加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算历史收益率均值
        method: 历史IC均值计算方法。'equal':算术平均；'half_life':半衰加权
        half_life: int, 半衰加权时的半衰期
    返回:
        DataFrame, 复合因子
    """
    # 计算各期IC
    ic_df = calc_ic(factor_df, mret_df, factor_list, return_col_name='nxt1_ret', ic_type='spearman')
    ic_df = ic_df.sort_values('trade_date')
    ic_df['trade_date'] = ic_df['trade_date'].shift(-1)
    
    # 计算历史因子收益率移动平均
    for fn in factor_list:
        if method == 'equal':
            ic_df[fn+'_ma'] = ic_df[fn].rolling(span).mean()
        elif method == 'half_life':
            ic_df[fn+'_ma'] = ic_df[fn].rolling(span).apply(lambda x: np.average(x, weights=list(reversed([0.5**(1.0*i/half_life) for i in range(span)]))))
        else:
            return 0
    
    ic_df = ic_df.dropna()
    ic_df = ic_df.drop(factor_list, axis=1)
    ic_df.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, ic_df)
    return conb_df, ic_df

def max_icir_combine(factor_df, mret_df, factor_list, span, method='sample', weight_limit=True):
    """
    最大化ICIR加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算IC均值和协方差矩阵
        method: 估计协方差矩阵的方法。'sample':直接用样本协方差矩阵；'shrunk':压缩估计
        weight_limit: bool, 是否约束权重为正
    返回:
        DataFrame, 复合因子
    """
    # 计算各期IC
    ic_df = calc_ic(factor_df, mret_df, factor_list, return_col_name='nxt1_ret', ic_type='spearman')
    ic_df = ic_df.sort_values('trade_date')
    ic_df['trade_date'] = ic_df['trade_date'].shift(-1)
    ic_df = ic_df.dropna()
    
    # 最大化ICIR
    m_ir_df = {}
    for i in range(span-1, len(ic_df)):
        dt = ic_df.ix[i, 'trade_date']
        ic_dt = ic_df.ix[i-span+1:i, factor_list]

        n = len(factor_list)
        # 求解最优化问题
        if method == 'sample':
            P = matrix(2*np.cov(ic_dt.T))
        elif method == 'shrunk':
            P = matrix(2*LedoitWolf().fit(ic_dt[factor_list].as_matrix()).covariance_)
        q = matrix([0.0]*n)
        G = matrix(-np.identity(n))
        h = matrix([0.0]*n)
        A = matrix(ic_dt.mean(), (1,n))
        b = matrix(1.0)
        if weight_limit:
            try:
                res = np.array(solvers.qp(P=P,q=q,G=G,h=h, A=A,b=b)['x'])
            except:
                res = np.array(solvers.qp(P=P,q=q, A=A,b=b)['x'])
        else:
            res = np.array(solvers.qp(P=P,q=q, A=A,b=b)['x'])

        m_ir_df[dt] = np.array(res).reshape(n)
    m_ir_df = pd.DataFrame(m_ir_df, index=factor_list).T.reset_index()
    if weight_limit:
        m_ir_df[factor_list] = np.where(m_ir_df[factor_list] < 0, 0, m_ir_df[factor_list])
    m_ir_df.loc[m_ir_df.sum(axis=1)==0, factor_list] = 1 
    m_ir_df.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, m_ir_df)
    return conb_df, m_ir_df

def max_ic_combine(factor_df, mret_df, factor_list, span, method='sample', weight_limit=True):
    """
    最大化IC加权法合成因子
    参数:
        factor_df: DataFrame, 待合成因子值
        mret_df: DataFrame, 个股收益率
        factor_list: list, 待合成因子列表
        span: 使用历史长度计算IC均值
        method: 估计协方差矩阵的方法。'sample':直接用样本协方差矩阵；'shrunk':压缩估计
        weight_limit: bool, 是否约束权重为正
    返回:
        DataFrame, 复合因子
    """
    # 计算各期IC
    ic_df = calc_ic(factor_df, mret_df, factor_list, return_col_name='nxt1_ret', ic_type='spearman')
    ic_df = ic_df.sort_values('trade_date')
    ic_df['trade_date'] = ic_df['trade_date'].shift(-1)
    for fn in factor_list:
        ic_df[fn] = ic_df[fn].rolling(span).mean()
    ic_df = ic_df.dropna()
    
    # 最大化IC
    m_ic_df = {}
    for dt in ic_df['trade_date']:
        ic_mean = ic_df.loc[ic_df['trade_date'] == dt, factor_list].values
        tmp_factor_df = factor_df.loc[factor_df['trade_date'] == dt, factor_list]

        n = len(factor_list)
        # 求解最优化问题
        if method == 'sample':
            P = matrix(2*np.cov(tmp_factor_df.T))
        elif method == 'shrunk':
            P = matrix(2*LedoitWolf().fit(tmp_factor_df.dropna().as_matrix()).covariance_)
        q = matrix([0.0]*n)
        G = matrix(-np.identity(n))
        h = matrix([0.0]*n)
        A = matrix(ic_mean, (1,n))
        b = matrix(1.0)
        if weight_limit:
            try:
                res = np.array(solvers.qp(P=P,q=q,G=G,h=h, A=A,b=b)['x'])
            except:
                res = np.array(solvers.qp(P=P,q=q, A=A,b=b)['x'])
        else:
            res = np.array(solvers.qp(P=P,q=q,A=A,b=b)['x'])

        m_ic_df[dt] = np.array(res).reshape(n)
    m_ic_df = pd.DataFrame(m_ic_df, index=factor_list).T.reset_index()
    if weight_limit:
        m_ic_df[factor_list] = np.where(m_ic_df[factor_list] < 0, 0, m_ic_df[factor_list])
    m_ic_df.loc[m_ic_df.sum(axis=1)==0, factor_list] = 1 
    m_ic_df.columns = ['trade_date']+factor_list
    
    # 因子加权
    conb_df = factor_combine(factor_df, factor_list, m_ic_df)
    return conb_df, m_ic_df


# -*- coding: utf-8 -*-
import pdb
from setuptools import setup
from setuptools import find_packages
from distutils.cmd import Command
from distutils.extension import Extension
import os
import sys
import io
import subprocess
import platform
import numpy as np
from Cython.Build import cythonize
from Cython.Distutils import build_ext
import Cython.Compiler.Options
Cython.Compiler.Options.annotate = True

if "--line_trace" in sys.argv:
    line_trace = True
    print("Build with line trace enabled ...")
    sys.argv.remove("--line_trace")
else:
    line_trace = False

PACKAGE = "ultron"
NAME = "Finance-Ultron"
VERSION = "0.4.2"
DESCRIPTION = "FinUltron " + VERSION
AUTHOR = "flaght"
AUTHOR_EMAIL = "flaght@gmail.com"
URL = 'https://github.com/flaght'


def git_version():
    from subprocess import Popen, PIPE
    gitproc = Popen(['git', 'rev-parse', 'HEAD'], stdout=PIPE)
    (stdout, _) = gitproc.communicate()
    return stdout.strip()

if platform.system() != "Windows":
    import multiprocessing
    n_cpu = multiprocessing.cpu_count()
else:
    n_cpu = 0

class version_build(Command):

    description = "test the distribution prior to install"

    user_options = [
        ('test-dir=', None,
         "directory that contains the test definitions"),
    ]

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        git_ver = git_version()[:10]
        configFile = 'ultron/__init__.py'

        file_handle = open(configFile, 'r')
        lines = file_handle.readlines()
        newFiles = []
        for line in lines:
            if line.startswith('__version__'):
                line = line.split('+')[0].rstrip()
                line = line + " + \"-" + git_ver + "\"\n"
            newFiles.append(line)
        file_handle.close()
        os.remove(configFile)
        file_handle = open(configFile, 'w')
        file_handle.writelines(newFiles)
        file_handle.close()

if sys.version_info > (3, 0, 0):
    requirements = "requirements/py3.txt"
else:
    requirements = "requirements/py2.txt"


ext_modules = [
    "ultron/sentry/Analysis/TechnicalAnalysis/StatefulTechnicalAnalysers.pyx",
    "ultron/sentry/Analysis/TechnicalAnalysis/StatelessTechnicalAnalysers.pyx",
    "ultron/sentry/Analysis/CrossSectionValueHolders.pyx",
    "ultron/sentry/Analysis/SecurityValueHolders.pyx",
    "ultron/sentry/Analysis/SeriesValues.pyx",
    "ultron/sentry/Analysis/transformer.pyx",
    "ultron/sentry/Math/Accumulators/IAccumulators.pyx",
    "ultron/sentry/Math/Accumulators/impl.pyx",
    "ultron/sentry/Math/Accumulators/StatefulAccumulators.pyx",
    "ultron/sentry/Math/Accumulators/StatelessAccumulators.pyx",
    "ultron/sentry/Math/Distributions/norm.pyx",
    "ultron/sentry/Math/Distributions/NormalDistribution.pyx",
    "ultron/sentry/Math/ErrorFunction.pyx",
    "ultron/sentry/Math/MathConstants.pyx",
    "ultron/sentry/Math/udfs.pyx",
    "ultron/sentry/Utilities/Asserts.pyx",
    "ultron/sentry/Utilities/Tools.pyx",
    "ultron/tradingday/Calendar.pyx",
    "ultron/tradingday/Date.pyx",
    "ultron/tradingday/Period.pyx",
    "ultron/tradingday/Schedule.pyx",
    "ultron/tradingday/Enums/TimeUnits.pyx",
    "ultron/tradingday/Enums/BizDayConventions.pyx",
    "ultron/tradingday/Enums/DateGeneration.pyx",
    "ultron/tradingday/Enums/Months.pyx",
    "ultron/tradingday/Enums/NormalizingType.pyx",
    "ultron/tradingday/Enums/OptionType.pyx",
    "ultron/tradingday/Enums/Weekdays.pyx"
]


def generate_extensions(ext_modules, line_trace=True):

    extensions = []

    if line_trace:
        print("define cython trace to True ...")
        define_macros = [('CYTHON_TRACE', 1), ('CYTHON_TRACE_NOGIL', 1)]
    else:
        define_macros = []

    for pyxfile in ext_modules:
        ext = Extension(name='.'.join(pyxfile.split('/'))[:-4],
                        sources=[pyxfile],
                        define_macros=define_macros)
        extensions.append(ext)
    return extensions


import multiprocessing
n_cpu = multiprocessing.cpu_count()

ext_modules_settings = cythonize(generate_extensions(ext_modules, line_trace), 
                                 compiler_directives={'embedsignature': True, 'linetrace': line_trace}, 
                                 nthreads=n_cpu)


setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    packages=find_packages(),
    include_package_data=False,
    install_requires=io.open(requirements, encoding='utf8').read(),
    classifiers=[],
    cmdclass={"version_build": version_build},
    ext_modules=ext_modules_settings,
    include_dirs=[np.get_include()],
)



# -*- coding: utf-8 -*-

from ultron.sentry.Analysis.SecurityValueHolders import SecurityShiftedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityDeltaValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityIIFValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityConstArrayValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSTopNSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSBottomNSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSAverageAdjustedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSFillNASecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityCurrentValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from ultron.sentry.Analysis import TechnicalAnalysis
from ultron.sentry.Analysis.transformer import transform

__all__ = ['SecurityShiftedValueHolder',
           'SecurityDeltaValueHolder',
           'SecurityIIFValueHolder',
           'SecurityConstArrayValueHolder',
           'CSRankedSecurityValueHolder',
           'CSTopNSecurityValueHolder',
           'CSBottomNSecurityValueHolder',
           'CSAverageSecurityValueHolder',
           'CSAverageAdjustedSecurityValueHolder',
           'CSZScoreSecurityValueHolder',
           'CSFillNASecurityValueHolder',
           'CSPercentileSecurityValueHolder',
           'CSResidueSecurityValueHolder',
           'SecurityCurrentValueHolder',
           'SecurityLatestValueHolder',
           'TechnicalAnalysis',
           'transform']
